#include "includes.h"

///////////////////////////////////////
// Settings
///////////////////////////////////////
#define FINITEFIELD FFGMP

#define FF31 0  // Field from p=2 to p=2^31-1
#define FF63 1  // Field from p=2 to p=2^63-1
#define FFGMP 2 // Field with arbitrary size


///////////////////////////////////////
// Types
///////////////////////////////////////
#if FINITEFIELD == FF31    
    typedef __int64_t mf_t;
#elif FINITEFIELD == FF63
    typedef __int128_t mf_t;
#elif FINITEFIELD == FFGMP
    typedef mpz_t mf_t;
#endif

typedef mf_t* mf_vec_t;


///////////////////////////////////////
// State
///////////////////////////////////////

extern mf_t mf_p;
extern size_t mf_p_bits;
extern size_t mf_p_bytes;

extern mf_t mf_zero;
extern mf_t mf_one;
extern mf_t mf_randmax;



///////////////////////////////////////
// Functions
///////////////////////////////////////

extern void mf_initialize(mf_t p);
extern void mf_setfield(mf_t p);
extern void mf_finalize();
extern void mf_print(mf_t x);
extern void mf_import(mf_t r, uint8_t* a);
extern void mf_export(uint8_t* r, mf_t a);
extern void mf_vec_import(mf_vec_t r, uint8_t* a, size_t l);
extern void mf_vec_export(uint8_t* r, mf_vec_t a, size_t l);




///////////////////////////////////////
// Finite field arithmetric
///////////////////////////////////////


#if (FINITEFIELD == FF31) || (FINITEFIELD == FF63)

    #define mf_init(r) { }
    #define mf_init0(r) { r = 0;}
    #define mf_init1(r) { }
    #define mf_init2(r) { }
    #define mf_clear(r) { }
    #define mf_string(r, s) { r = strtoll(s, NULL, 0); }
    #define mf_int(r, d) { r = (mf_t)d; }
    #define mf_set(r, a) { r = a; }
    #define mf_set_ui(r, a) { r = (mf_t)a; }
    #define mf_add(r, a, b) { r = ((a+b) > mf_p ? (a+b-mf_p) : (a+b)); };
    #define mf_sub(r, a, b) { r = ((a-b) < 0    ? (a-b+mf_p) : (a-b)); };
    #define mf_mul(r, a, b) { r = (a*b) % mf_p; };
    extern void mf_pow(mf_t* r, mf_t u);
    extern void mf_inv(mf_t* r, mf_t u);
    #define mf_swap(r, a) { a = r+a; r = a-r; a = a-r; }
    #define mf_iszero(r) (r == 0)
    extern void mf_srand(mf_t* r, mprg_t* prg);
    #define mf_srandn0(r, p) { do { r = mf_srandn0(r, p); } while(r == 0); }
#if FINITEFIELD == FF31
    #define mf_rand(r) { r = (mf_t)sfmt_genrand_uint32(&mf_sfmt_randstate) % mf_p; }
#else
    #define mf_rand(r) { r = (mf_t)sfmt_genrand_uint64(&mf_sfmt_randstate) % mf_p; }
#endif

#elif FINITEFIELD == FFGMP

    #define mf_init(r) { mpz_init(r); }
    #define mf_init0(r) { mpz_init(r); }
    #define mf_init1(r) { mpz_init2(r, mf_p_bits); }
    #define mf_init2(r) { mpz_init2(r, mf_p_bits*2); }
    #define mf_clear(r) { mpz_clear(r); }
    #define mf_string(r, s) { mpz_set_str(r, s, 10); }
    #define mf_int(r, d) { mpz_set_ui(r, d); }
    #define mf_set(r, a) { mpz_set(r, a); }
    #define mf_set_ui(r, a) { mpz_set_ui(r, a); }
    #define mf_add(r, a, b) { mpz_add(r, a, b); if(mpz_cmp(r, mf_p) >= 0) mpz_sub(r, r, mf_p); }
    #define mf_sub(r, a, b) { mpz_sub(r, a, b); if(mpz_cmp(r, mf_zero) < 0) mpz_add(r, r, mf_p); }
    #define mf_mul(r, a, b) { mpz_mul(r, a, b); mpz_mod(r, r, mf_p); }
    #define mf_pow(r, a, b) { mpz_powm(r, a, b, mf_p); }
    #define mf_inv(r, a) { mpz_invert(r, a, mf_p); }
    #define mf_swap(r, a) { mpz_swap(r, a); }
    #define mf_iszero(r) (mpz_cmp(mf_zero, r) == 0)
    #define mf_rand(r) { mpz_urandomm(r, mf_gmp_randstate, mf_p); }
    extern void mf_srand(mf_t r, mprg_t* prg);
    #define mf_srandn0(r, p) { do { mf_srand(r, p); } while(mpz_cmp(mf_zero, r) == 0); }
    
#endif




///////////////////////////////////////
// Vector arithmetric
///////////////////////////////////////

#define mf_vec_alloc(r, l) { r = malloc(l * sizeof(mf_t)); }
#define mf_vec_clear(r, l) { for(int tmpI=0; tmpI<l; ++tmpI) { mf_clear(r[tmpI]); } free(r); }
#define mf_vec_init(r, l) { mf_vec_alloc(r, l); for(int tmpI=0; tmpI<l; ++tmpI) { mf_init(r[tmpI]); } }
#define mf_vec_init0(r, l) { mf_vec_alloc(r, l); for(int tmpI=0; tmpI<l; ++tmpI) { mf_init0(r[tmpI]); } }
#define mf_vec_init1(r, l) { mf_vec_alloc(r, l); for(int tmpI=0; tmpI<l; ++tmpI) { mf_init1(r[tmpI]); } }
#define mf_vec_init2(r, l) { mf_vec_alloc(r, l); for(int tmpI=0; tmpI<l; ++tmpI) { mf_init2(r[tmpI]); } }
#define mf_vec_rand(r, l) { for(int tmpI=0; tmpI<l; ++tmpI) { mf_rand(r[tmpI]); } }
#define mf_vec_srand(r, l, p) { for(int tmpI=0; tmpI<l; ++tmpI) { mf_srand(r[tmpI], p); } }
#define mf_vec_srandn0(r, l, p) { for(int tmpI=0; tmpI<l; ++tmpI) { mf_srandn0(r[tmpI], p); } }
#define mf_vec_zero(r, l) { for(int tmpI=0; tmpI<l; ++tmpI) { mf_set(r[tmpI], mf_zero); } }
#define mf_vec_add(r, a, b, l) { for(int tmpI=0; tmpI<l; ++tmpI) { mf_add(r[tmpI], a[tmpI], b[tmpI]); } }
#define mf_vec_sub(r, a, b, l) { for(int tmpI=0; tmpI<l; ++tmpI) { mf_sub(r[tmpI], a[tmpI], b[tmpI]); } }
#define mf_vec_mul(r, a, b, l) { for(int tmpI=0; tmpI<l; ++tmpI) { mf_mul(r[tmpI], a[tmpI], b[tmpI]); } }
#define mf_vec_scale(r, a, b, l) { for(int tmpI=0; tmpI<l; ++tmpI) { mf_mul(r[tmpI], a[tmpI], b); } }