#include "io.h"

///////////////////////////////////////
// WSA for Windows compatibility
///////////////////////////////////////

#ifdef WIN32
WSADATA wsa;
void close(sock_t sock) { closesocket(sock); }
#else
int WSAGetLastError() { return errno; }
#endif


///////////////////////////////////////
// Initialization and destruction
///////////////////////////////////////

void mio_initialize() {
	clientctx = NULL;
	serverctx = NULL;

	init_openssl();

	#ifdef WIN32
	if (WSAStartup(MAKEWORD(2, 2), &wsa) != 0) {
		printf("Could not start WSA\n");
		exit(1);
	}
	#endif
}

void mio_finalize() {
	cleanup_openssl();
	#ifdef WIN32
		WSACleanup();
	#endif
}

void mio_close(sock_t sock) {
	close(sock);
}

void mio_check(sock_t sock) {
	uint8_t val = time(NULL);
	uint8_t ping[1];
	ping[0] = val;
	printf("Now pinging with %d\n", val);
	mio_send(sock, 1, ping);
	mio_recv(sock, 1, ping);
	if (ping[0] != val) {
		printf("Invalid pong from server\n");
		exit(1);
	}
	else {
		printf("Pong was good!\n");
	}
}

void mio_parseaddr(sockaddrp_t addr, char* ip, int port) {
	memset(addr, 0, sizeof(sockaddr_t));
	addr->sin_family = AF_INET;
	addr->sin_port = htons(port);
	addr->sin_addr.s_addr = inet_addr(ip);
}







///////////////////////////////////////
// Plain server
///////////////////////////////////////

sock_t mio_listen(int port) {
	return create_socket(port);
}

sock_t mio_accept(sock_t sock, sockaddrp_t addr) {
	socklen_t len = sizeof(sockaddr_t);
	sock_t client = accept(sock, (struct sockaddr*)addr, &len);
	if(client < 0) {
		printf("Could not accept\n");
		exit(1);
	}
	int flag = ENABLEENAGLE;
	setsockopt(sock, IPPROTO_TCP, TCP_NODELAY, (char *)&flag, sizeof(int));
	return client;
}

sock_t mio_connect(sockaddrp_t addr) {
	socklen_t len = sizeof(sockaddr_t);
	sock_t sock = socket(AF_INET, SOCK_STREAM, 0);
	if (sock == -1)
	{
		printf("Could not create socket\n");
		exit(1);
	}
	if (connect(sock, (struct sockaddr*)addr, len) < 0) {
		printf("Could not connect %s\n", strerror(errno));
		exit(1);
	}
	int flag = ENABLEENAGLE;
	setsockopt(sock, IPPROTO_TCP, TCP_NODELAY, (char *)&flag, sizeof(int));
	return sock;
}

int mio_send(sock_t sock, size_t len, uint8_t* buffer) {
	int p = 0;
	int sent;
	while(p < len) {
		sent = send(sock, (void*)&buffer[p], len-p, 0);
		if (sent == -1) {
			printf("Could not send %i %i %i: %s\n", WSAGetLastError(), p, sent, strerror(errno));
			return -1;
		}
		p += sent;
	}
	return p;
}

int mio_recv(sock_t sock, size_t len, uint8_t* buffer) {
	int p = 0;
	int recvd;
	while(p < len) {
		recvd = recv(sock, &buffer[p], len-p, 0);
		if (recvd == -1) {
			printf("Could not read %i %i %i %li %li\n", WSAGetLastError(), p, recvd, len-p, len);
			return -1;
		}
		p += recvd;
	}
	return p;
}




///////////////////////////////////////
// Secure server
///////////////////////////////////////

sock_t mio_saccept(sock_t sock, SSL** ssl, sockaddrp_t addr)
{
	socklen_t len = sizeof(sockaddr_t);
	sock_t client = accept(sock, (struct sockaddr*)addr, &len);
	if(sock < 0) {
		printf("Could not accept\n");
		return -1;
	}
	int flag = ENABLEENAGLE;
	setsockopt(sock, IPPROTO_TCP, TCP_NODELAY, (char *)&flag, sizeof(int));

	*ssl = SSL_new(serverctx);
	if(*ssl == NULL) {
		printf("Could not create SSL!");
		return -1;
	}
	SSL_set_fd(*ssl, client);
	int accept = SSL_accept(*ssl);
	if (accept <= 0) {
		printf("Could not accept SSL\n");
		ERR_print_errors_fp(stderr);
		int errcode = SSL_get_error(*ssl, accept);
		switch(errcode)
		{
			case SSL_ERROR_NONE: break;
			case SSL_ERROR_ZERO_RETURN: printf("SSL connect returned 0.\n");break;
			case SSL_ERROR_WANT_READ: printf("SSL connect: Read Error.\n");break;
			case SSL_ERROR_WANT_WRITE: printf("SSL connect: Write Error.\n");break;
			case SSL_ERROR_WANT_CONNECT: printf("SSL connect: Error connect.\n"); break;
			case SSL_ERROR_WANT_ACCEPT: printf("SSL connect: Error accept.\n"); break;
			case SSL_ERROR_WANT_X509_LOOKUP: printf("SSL connect error: X509 lookup.\n"); break;
			case SSL_ERROR_SYSCALL: printf("SSL connect: Error in system call.\n"); break;
			case SSL_ERROR_SSL: printf("SSL connect: Protocol Error.\n"); break;
			default: printf("Failed SSL connect.\n");
		}
		return -1;
	}

	return client;
}

sock_t mio_sconnect(SSL** ssl, sockaddrp_t addr) {
	socklen_t len = sizeof(sockaddr_t);
	sock_t sock = socket(AF_INET, SOCK_STREAM, 0);
	if (sock == -1)
	{
		printf("Could not create socket\n");
		exit(1);
	}
	if (connect(sock, (struct sockaddr*)addr, len) < 0) {
		printf("Could not connect %s\n", strerror(errno));
		exit(1);
	}
	int flag = ENABLEENAGLE;
	setsockopt(sock, IPPROTO_TCP, TCP_NODELAY, (char *)&flag, sizeof(int));

	*ssl = SSL_new(clientctx);
	if(*ssl == NULL) {
		printf("Could not create SSL!");
	}
	SSL_set_fd(*ssl, sock);
	int connect = SSL_connect(*ssl);
	if (connect <= 0)
	{
		int errcode = SSL_get_error(*ssl, connect);
		switch(errcode)
		{
			case SSL_ERROR_NONE: break;
			case SSL_ERROR_ZERO_RETURN: printf("SSL connect returned 0.\n");break;
			case SSL_ERROR_WANT_READ: printf("SSL connect: Read Error.\n");break;
			case SSL_ERROR_WANT_WRITE: printf("SSL connect: Write Error.\n");break;
			case SSL_ERROR_WANT_CONNECT: printf("SSL connect: Error connect.\n"); break;
			case SSL_ERROR_WANT_ACCEPT: printf("SSL connect: Error accept.\n"); break;
			case SSL_ERROR_WANT_X509_LOOKUP: printf("SSL connect error: X509 lookup.\n"); break;
			case SSL_ERROR_SYSCALL: printf("SSL connect: Error in system call.\n"); break;
			case SSL_ERROR_SSL: printf("SSL connect: Protocol Error.\n"); break;
			default: printf("Failed SSL connect.\n");
		}
		return -1;
	}

	return sock;
}

int mio_ssend(SSL* ssl, size_t len, uint8_t* buffer) {
	int p = 0;
	int sent;
	while(p < len) {
		sent = SSL_write(ssl, &buffer[p], len-p);
		if (sent <= 0) {
			printf("Could not send %i %i %i: %s\n", WSAGetLastError(), p, sent, strerror(errno));
			int errcode = SSL_get_error(ssl, sent);
			switch(errcode)
			{
				case SSL_ERROR_NONE: break;
				case SSL_ERROR_ZERO_RETURN: printf("SSL connect returned 0.\n");break;
				case SSL_ERROR_WANT_READ: printf("SSL connect: Read Error.\n");break;
				case SSL_ERROR_WANT_WRITE: printf("SSL connect: Write Error.\n");break;
				case SSL_ERROR_WANT_CONNECT: printf("SSL connect: Error connect.\n"); break;
				case SSL_ERROR_WANT_ACCEPT: printf("SSL connect: Error accept.\n"); break;
				case SSL_ERROR_WANT_X509_LOOKUP: printf("SSL connect error: X509 lookup.\n"); break;
				case SSL_ERROR_SYSCALL: printf("SSL connect: Error in system call.\n"); break;
				case SSL_ERROR_SSL: printf("SSL connect: Protocol Error.\n"); break;
				default: printf("Failed SSL connect.\n");
			}
			return -1;
		}
		p += sent;
	}
	return p;
}

int mio_srecv(SSL* ssl, size_t len, uint8_t* buffer) {
	int p = 0;
	int recvd;
	while(p < len) {
		recvd = SSL_read(ssl, &buffer[p], len-p);
		if (recvd <= 0) {
			printf("Could not read %i %i %i: %s\n", WSAGetLastError(), p, recvd, strerror(errno));
			int errcode = SSL_get_error(ssl, recvd);
			switch(errcode)
			{
				case SSL_ERROR_NONE: break;
				case SSL_ERROR_ZERO_RETURN: printf("SSL connect returned 0.\n");break;
				case SSL_ERROR_WANT_READ: printf("SSL connect: Read Error.\n");break;
				case SSL_ERROR_WANT_WRITE: printf("SSL connect: Write Error.\n");break;
				case SSL_ERROR_WANT_CONNECT: printf("SSL connect: Error connect.\n"); break;
				case SSL_ERROR_WANT_ACCEPT: printf("SSL connect: Error accept.\n"); break;
				case SSL_ERROR_WANT_X509_LOOKUP: printf("SSL connect error: X509 lookup.\n"); break;
				case SSL_ERROR_SYSCALL: printf("SSL connect: Error in system call.\n"); break;
				case SSL_ERROR_SSL: printf("SSL connect: Protocol Error.\n"); break;
				default: printf("Failed SSL connect.\n");
			}
			return -1;
		}
		p += recvd;
	}
	return p;
}

void mio_sslfree(SSL* ssl) {
    SSL_shutdown(ssl);
	SSL_free(ssl);
}







///////////////////////////////////////
// Helpers
///////////////////////////////////////

int create_socket(int port)
{
	int s;

    sockaddr_t addr;
    addr.sin_family = AF_INET;
    addr.sin_port = htons(port);
    addr.sin_addr.s_addr = htonl(INADDR_ANY);

    s = socket(AF_INET, SOCK_STREAM, 0);
    if (s < 0) {
		printf("Unable to create socket");
		exit(1);
    };

    if (bind(s, (struct sockaddr*)&addr, sizeof(addr)) < 0) {
		printf("Unable to bind");
		exit(1);
    }

    if (listen(s, 5) < 0) {
		printf("Unable to listen");
		exit(1);
    }

    return s;
}

void init_openssl()
{
	SSL_library_init();
    SSL_load_error_strings();
	ERR_load_BIO_strings();
	ERR_load_crypto_strings();
    OpenSSL_add_all_algorithms();

	create_client_context();
	create_server_context();
	configure_server_context();
	configure_client_context();
}

void cleanup_openssl()
{
	ERR_free_strings();
	EVP_cleanup();
	sk_SSL_COMP_free(SSL_COMP_get_compression_methods());
	CRYPTO_cleanup_all_ex_data();
}

void create_server_context()
{
    const SSL_METHOD *method = SSLv23_server_method();
    serverctx = SSL_CTX_new(method);
    if (!serverctx) {
		perror("Unable to create SSL context");
		ERR_print_errors_fp(stderr);
		exit(1);
    }
}

void create_client_context()
{
    const SSL_METHOD *method = SSLv23_client_method();
    clientctx = SSL_CTX_new(method);
    if (!clientctx) {
		perror("Unable to create SSL context");
		ERR_print_errors_fp(stderr);
		exit(1);
    }
}

void configure_server_context()
{
	SSL_CTX_set_ecdh_auto(serverctx, 1);

	if(SSL_CTX_use_certificate_chain_file(serverctx, "serverchain.pem") <= 0) {
        ERR_print_errors_fp(stderr);
		exit(1);
	}

    if (SSL_CTX_use_PrivateKey_file(serverctx, "server.key", SSL_FILETYPE_PEM) <= 0 ) {
        ERR_print_errors_fp(stderr);
		exit(1);
    }
}

void configure_client_context()
{
	SSL_CTX_set_verify_depth(clientctx, 4);
	SSL_CTX_set_verify(clientctx, SSL_VERIFY_PEER, NULL);
	SSL_CTX_load_verify_locations(clientctx, "root-ca.crt", NULL);
}
