#include "server.h"


static void* handleClient(void* c);

typedef struct {
	sock_t client;
	SSL* ssl;
	pthread_t thread;
} conn_t;

uint8_t* key;


void runServer() {
	printf("Running secure server on port %d\n", parms.port);

	mprg_t* prg = malloc(sizeof(mprg_t));
	mprg_init(prg);
	__uint128_t rand = mprg_rand128(prg);
	key = malloc(16);
	memcpy(key, &rand, 16);
	free(prg);
	printf("Secret key: "); printHex(key, 16); printf("\n");

	printf("Waiting for connections...\n");
	sock_t server = mio_listen(parms.port);

	conn_t* conns = malloc(sizeof(conn_t) * MAXSERVERTHREADS);
	int index = 0;
	while(true) {
		conn_t* conn = &conns[index];

		sockaddr_t addr;
		memset(&addr, 0, sizeof(sockaddr_t));

		conn->client = mio_saccept(server, &conn->ssl, &addr);
		if(conn->client == -1) {
			printf("Could not connect to client\n");
			continue;
		}

		char *clientip = inet_ntoa(addr.sin_addr);
		printf("New connection from %s\n", clientip);

		pthread_create(&conn->thread, NULL, handleClient, (void*)conn);

		index = (index+1) % MAXSERVERTHREADS;
	}

	printf("Server done\n");

	free(conns);

	mio_close(server);
	mio_finalize();
}






static void* handleClient(void* c) {
	conn_t* conn = (conn_t*)c;
	sock_t client = conn->client;
	SSL* ssl = conn->ssl;
	int recvd = 0;




	////////////////////////////////////////
	// Get field, note: benchmarking only, unsafe!
	////////////////////////////////////////
	uint8_t* fieldbuffer = malloc(MAXFIELDBYTES);
	uint32_t fieldbytes;
	recvd = mio_srecv(ssl, 4, (uint8_t*)&fieldbytes);
	fieldbytes = ntohl(fieldbytes);
	if(recvd <= 0) {
		printf("Could not receive field size.\n");
		free(fieldbuffer);
		goto endHandleClient;
	}
	if(fieldbytes > MAXFIELDBYTES) {
		printf("Client requested too large field.\n");
		free(fieldbuffer);
		goto endHandleClient;
	}
	recvd = mio_srecv(ssl, fieldbytes, fieldbuffer);
	if(recvd <= 0) {
		printf("Client requested invalid field!\n");
		free(fieldbuffer);
		goto endHandleClient;
	}
	mf_t p; mf_init(p);
	mpz_import(p, 1, 1, fieldbytes, 0, 0, fieldbuffer);
	mf_setfield(p);
	free(fieldbuffer);
	gmp_printf("Field chosen is %d-bytes = %d-bits and is: %Zd\n", mf_p_bytes, mf_p_bits, mf_p);




	////////////////////////////////////////
	// Authenticate
	////////////////////////////////////////
	uint8_t* authbuffer = malloc(auth_export_bytes);
	int canread = mio_srecv(ssl, auth_export_bytes, authbuffer);
	if(canread <= 0) {
		printf("Could not receive authentication.\n");
		free(authbuffer);
		goto endHandleClient;
	}
	uint8_t isAlice = authbuffer[0];
	auth_t auth = auth_import(authbuffer);
	if(!auth_verify(auth)) {
		printf("Authentication was invalid.\n");
		free(authbuffer);
		goto endHandleClient;
	}
	else {
		printf("Authentication good.\n");
	}




	////////////////////////////////////////
	// Generate randomness
	////////////////////////////////////////
	unsigned char hash[SHA256_DIGEST_LENGTH];
	SHA256_CTX sha256;
	SHA256_Init(&sha256);
	SHA256_Update(&sha256, key, 16);
	SHA256_Update(&sha256, &authbuffer[1], 3*auth_bytes);
	SHA256_Final(hash, &sha256);
	free(authbuffer);
	printf("Using PRG seed: "); printHex(hash, 16); printf("\n");

    mprg_t* prg = malloc(sizeof(mprg_t));
	mprg_seed(prg, hash);



	////////////////////////////////////////
	// Finite number of triples
	////////////////////////////////////////
	mf_vec_t a, b, c1, c2;
	mf_vec_init1(a, MAXTRIPLES);
	mf_vec_init1(b, MAXTRIPLES);
	mf_vec_init1(c1, MAXTRIPLES);
	mf_vec_init2(c2, MAXTRIPLES);

	size_t buffersize = 2 * MAXTRIPLES * mf_p_bytes;
	uint8_t* buffer = malloc(buffersize);

	uint32_t n;
	while(true) {
		recvd = mio_srecv(ssl, 4, (uint8_t*)&n);
		if(recvd <= 0) {
			printf("Disconnected client: Could not receive triple request.\n");
			break;
		}

		n = ntohl(n);
		if(n > MAXTRIPLES) {
			printf("Disconnected client: Client requested too many triples.\n");
			break;
		}

		// Construct triples
		mf_vec_srandn0(a, n, prg);
		mf_vec_srandn0(b, n, prg);
		mf_vec_srandn0(c1, n, prg);
		if(!isAlice) {
			mf_vec_mul(c2, a, b, n);
			mf_vec_sub(c2, c2, c1, n);
		}

		// Export triples
		if(isAlice) {
			mf_vec_export(&buffer[0], a, n);
			mf_vec_export(&buffer[n * mf_p_bytes], c1, n);
		}
		else {
			mf_vec_export(&buffer[0], b, n);
			mf_vec_export(&buffer[n * mf_p_bytes], c2, n);
		}

		// Send
		int sent = mio_ssend(ssl, n * 2 * mf_p_bytes, buffer);
		if(sent <= 0) {
			printf("Disconnected client: Could not send triples.\n");
			break;
		}

	}

	free(buffer);
	free(prg);





	////////////////////////////////////////
	// Clean up
	////////////////////////////////////////

endHandleClient:
	mio_sslfree(ssl);
	mio_close(client);

	return 0;
}
